/* Implementation of the TLMimeItem class.
   This file is part of TL, Tiggr's Library.
   Written by Tiggr <tiggr@es.ele.tue.nl>
   Copyright (C) 1995, 1996 Pieter J. Schoenmakers
   TL is distributed WITHOUT ANY WARRANTY.
   See the file LICENSE in the TL distribution for details.

   $Id: TLMimeItem.m,v 1.11 1996/01/23 12:18:34 tiggr Exp $  */

#define TLMIMEITEM_DECLARE_PRIVATE_METHODS

#import "tl/TLMimeItem.h"
#import "tl/TLPatchedRoots.h"
#import "tl/TLMutableString.h"
#import "tl/TLCons.h"
#import "tl/support.h"
#import <stdlib.h>
#import <ctype.h>

@implementation TLMimeItem

+(TLMimeItem *) mimeItemWithHeaders: (id <TLString>) h range: (id <TLRange>) r
{
  return [[self gcAlloc] initWithHeaders: h range: r];
} /* +mimeItemWithHeaders:range: */

-(void) addHeader: (TLString *) key value: (TLString *) val
{
  /* Cache the values of well-known important headers.  */
  if ([key strcasecmp: @"content-type"])
    {
      TLRange *slash = [val rangeOfString: @"/"];
      int end = [val indexOfCharacterMatchFrom: @" \t;"];

      if (content_type)
	[self warning: "duplicate content type ignored: %#", val];
      else
	{
	  content_type = [val stringWithRange: [TLRange rangeWithStart: 0
							end: [slash _start]]];
	  content_format = [val stringWithRange:
				  [TLRange rangeWithStart: [slash _end]
					   end: end]];
	  if (end != -1)
	    content_options = [val stringWithRange:
				     [TLRange rangeWithStart: end length: -1]];
	}
    }
  else if ([key strcasecmp: @"content-length"])
    content_length = [TLNumber numberWithInt: atoi ([val cString])];
  else if ([key strcasecmp: @"mime-version"])
    mime_version = val;
  else if ([key strcasecmp: @"content-transfer-encoding"])
    xfer_encoding = val;

  [headers addElement: CONS (key, val)];
} /* -addHeader:value: */

-(id <TLString>) header: (id <TLString>) h
{
  id <TLEnumerator> e = [headers enumerator];
  TLString *s;
  TLCons *c;

  /* Be paranoid.  */
  if (h)
    {
      s = [h string];
      while ((c = [e nextObject]))
	if ([s strcasecmp: [c car]])
	  return [c cdr];
    }

  return nil;
} /* -header: */

-initWithHeaders: (id <TLString>) hdr range: (id <TLRange>) r
{
  const char *s = [hdr cString] + [r _start];
  int start = 0, len = [r length];
  TLMutableString *value;
  TLString *header;

  headers = [TLVector vector];
  while (start < len)
    {
      int h;

      for (h = start; h < len && s[h] != ':' && !isspace (s[h]); h++);
      if (h == len || s[h] != ':')
	{
	  [self warning: "malformed header at position %d: `%#'", start, hdr];
	  INIT_RETURN_NIL ();
	}

      header = [TLString stringWithCString: s + start length: h - start];
      value = nil;
      h++;
      do
	{
	  for (start = h; start < len && (s[start] == ' ' || s[start] == '\t');
	       start++);

	  for (h = start; h < len && s[h] != '\r' && s[h] != '\n'; h++);
	  if (value)
	    [value appendCChar: ' '];
	  else
	    value = [TLMutableString mutableString];
	  [value appendCString: s + start length: h - start];

	  /* The following code implies a header field is terminated by
	     the regexp `\r*\n?' with length > 0.  This is needed for a
	     case spotted with: `User-Agent: OmniWeb/1.0 libwww/2.16pre2
	     via proxy gateway CERN-HTTPD/3.0 libwww/2.17'.  */
	  while (h < len && s[h] == '\r')
	    h++;
	  if (h < len && s[h] == '\n')
	    h++;

	  start = h;
	} while (start < len && (s[start] == ' ' || s[start] == '\t'));

      [self addHeader: header value: value];
    }

  return self;
} /* -initWithHeaders:range: */

-contents
{
  return (contents);
} /* -contents */

-(TLNumber *) contentLength
{
  return (content_length);
} /* -contentLength */

-(TLString *) contentType
{
  return (content_type);
} /* -contentType */

-(TLString *) contentFormat
{
  return (content_format);
} /* -contentFormat */

-(TLString *) contentTypeOptions
{
  return (content_options);
} /* -contentTypeOptions */

-(TLString *) contentTransferEncoding
{
  return (xfer_encoding);
} /* -contentTransferEncoding */

-(void) gcReference
{
  MARK (headers);
  MARK (content_type);
  MARK (content_format);
  MARK (content_options);
  MARK (content_length);
  MARK (xfer_encoding);
  MARK (mime_version);
  MARK (contents);
} /* -gcReference */

-(id <TLVector>) headers
{
  return (headers);
} /* -headers */

-(TLString *) mimeVersion
{
  return (mime_version);
} /* -mimeVersion */

@end
