/* Implementation of TLRunLoop class.
   This file is part of TL, Tiggr's Library.
   Written by Tiggr <tiggr@es.ele.tue.nl>
   Copyright (C) 1995, 1996 Pieter J. Schoenmakers
   TL is distributed WITHOUT ANY WARRANTY.
   See the file LICENSE in the TL distribution for details.

   $Id: TLRunLoop.m,v 1.13 1996/01/23 12:18:38 tiggr Exp $  */

#define TLRUNLOOP_DECLARE_PRIVATE_METHODS
#import "tl/support.h"
#import "tl/TLPatchedRoots.h"
#import "tl/TLRunLoop.h"
#import "tl/TLGC.h"
#import <string.h>
#if HAVE_STRINGS_H
#import <strings.h>
#endif

@implementation TLRunLoop

+(TLRunLoop *) runLoop
{
  return ([[self gcAlloc] init]);
} /* +runLoop */

+(TLRunLoop *) sharedRunLoop
{
  static id shared;

  if (!shared)
    {
      shared = [self runLoop];
      [shared gcLock];
    }
  return (shared);
} /* +sharedRunLoop */

-(void) gcMark
{
  [super gcMark];
} /* -gcMark */

-init
{
  return (self);
} /* -init */

-addStream: (id <TLStream>) stream delegate: delegate toSet: (int) idx
{
  struct tlrunloop_fd_info *fdi;
  int fd;
  
  if (!stream || idx < 0 || idx > 2)
    return (nil);

  fd = [stream fileDescriptor];
  if (fd < 0)
    return (nil);

  fdi = &fds[idx];

  if (fd >= fdi->num)
    {
      changed = 1;
      fdi->num = fd + 1;
    }

  if (fd >= fdi->cap)
    {
      fdi->streams = xrealloc (fdi->streams, (fd + 1) * sizeof (*fdi->streams));
      bzero ((void *) (fdi->streams + fdi->cap),
	     (fd + 1 - fdi->cap) * sizeof (*fdi->streams));
      fdi->delegates = xrealloc (fdi->delegates,
				 (fd + 1) * sizeof (*fdi->delegates));
      bzero ((void *) (fdi->delegates + fdi->cap),
	     (fd + 1 - fdi->cap) * sizeof (*fdi->delegates));
      fdi->cap = fd + 1;
    }

  FD_SET (fd, &fdi->set);
  ASGN_IVAR (fdi->streams[fd], stream);
  ASGN_IVAR (fdi->delegates[fd], delegate ? delegate : stream);

  return (stream);
} /* -addStream:delegate:idx: */

-addStreamForRead: (id <TLStream>) stream
 delegate: (id <TLRunLoopReadDelegate>) delegate
{
  return ([self addStream: stream delegate: delegate toSet: 0]);
} /* -addStreamForRead:delegate: */

-addStreamForWrite: (id <TLStream>) stream
 delegate: (id <TLRunLoopWriteDelegate>) delegate
{
  return ([self addStream: stream delegate: delegate toSet: 1]);
} /* -addStreamForWrite:delegate: */

-addStreamForException: (id <TLStream>) stream
 delegate: (id <TLRunLoopExceptionDelegate>) delegate
{
  return ([self addStream: stream delegate: delegate toSet: 2]);
} /* -addStreamForException:delegate: */

-(TLDate *) inactivityTimeout
{
  return (inactive_timeout);
} /* -inactivityTimeout */

-(TLDate *) inactivityGCTimeLimit
{
  return (inactive_gc_time_limit);
} /* -inactivityGCTimeLimit */

-(unsigned long) inactivityPartialThreshold
{
  return (inactive_gc_partial_threshold);
} /* -inactivityPartialThreshold */

-(unsigned long) inactivityTotalThreshold
{
  return (inactive_gc_total_threshold);
} /* -inactivityTotalThreshold */

-removeStream: (id <TLStream>) stream fromSet: (int) idx
{
  struct tlrunloop_fd_info *fdi;
  int fd;
  
  if (!stream || idx < 0 || idx > 2)
    return (nil);

  fd = [stream fileDescriptor];
  if (fd < 0)
    return (nil);

  fdi = &fds[idx];
  if (fd >= fdi->num)
    return (nil);

  FD_CLR (fd, &fdi->set);
  fdi->streams[idx] = nil;
  fdi->delegates[idx] = nil;

  if (fd + 1 == fdi->num)
    {
      changed = 1;
      for (fdi->num--; fdi->num >= 0; fdi->num--)
	if (FD_ISSET (fdi->num, &fdi->set))
	  break;
      fdi->num++;
    }
  return (stream);
} /* -removeStream:fromSet: */

-removeStreamForRead: (id <TLStream>) s
{
  return ([self removeStream: s fromSet: 0]);
} /* -removeStreamForRead: */

-removeStreamForWrite: (id <TLStream>) s
{
  return ([self removeStream: s fromSet: 1]);
} /* -removeStreamForWrite: */

-removeStreamForException: (id <TLStream>) s
{
  return ([self removeStream: s fromSet: 2]);
} /* -removeStreamForException: */

-run
{
  return ([self runUntilDate: [TLDate distantFuture]]);
} /* -run */

-runUntilDate: (TLDate *) limit
{
  int i, n, set, num = 0, do_gc;
  fd_set cfds[3];
  GCDECL1;

  GCPRO1 (limit);
  changed = 1;

  /* Clear unnecessary bits, for easy debugging.  */
  bzero ((void *) cfds, sizeof (cfds));

  do
    {
      time_interval wait;
      
      if (changed)
	{
	  changed = 0;
	  for (num = set = 0; set < 3; set++)
	    if (fds[set].num > num)
	      num = fds[set].num;
	}

      for (set = 0; set < 3; set++)
	memcpy (&cfds[set], &fds[set].set, sizeof (fds[set]).set);

      do_gc = (inactive_gc_time_limit
	       && ((inactive_gc_partial_threshold
		    && inactive_gc_partial_threshold < tlgc_alloc_since_partial)
		   || (inactive_gc_total_threshold
		       && (inactive_gc_total_threshold
			   < TLGC_TOTAL_ALLOC_SINCE_COMPLETE))));

      if (!limit || stop)
	wait = 0;
      else
	{
	  time_interval now = time_since_unix_epoch ();

	  wait = [limit interval];
	  if (wait < now)
	    stop = 1, wait = 0;
	  else
	    wait -= now;
	}

      /* If we want to timeout for GC and have a limit, make WAIT the
         minimum of both.  */
      if (do_gc && !stop)
	{
	  time_interval gctv = [inactive_timeout interval];

	  if (gctv < wait)
	    wait = gctv;
	}

      {
	struct timeval tv = timeval_from_time_interval (wait);

#ifdef __NeXT__
      /* NeXT's select(2) is braindead concerning the timeout argument.
         Patch it.  */
	if ((unsigned) tv.tv_sec > 86400)
	  tv.tv_sec = 86400;
#endif

	n = select (num, (void *) &cfds[0], (void *) &cfds[1],
		    (void *) &cfds[2], &tv);
      }

      if (!n)
	if (do_gc)
	  tl_garbage_collect ([inactive_gc_time_limit microSeconds]);
	else
	  {
#ifndef __NeXT__
	    /* Do not issue a warning on the NeXT, where it can occur
               often (i.e. once a day).  */
	    [self warning: "no descriptors from select"];
#endif
	  }
      else if (n < 0)
	{
	  if (ERRNO != EINTR)
	    [self error: "select: %s", ERRMSG];
	}
      else for (set = 0; n && set < 3; set++)
	for (i = 0; n && i < fds[set].num; i++)
	  if (FD_ISSET (i, &cfds[set]))
	    {
	      id d = fds[set].delegates[i];
	      id s = fds[set].streams[i];

	      n--;
	      switch (set)
		{
		case 0: [d readEventOnStream: s]; break;
		case 1: [d writeEventOnStream: s]; break;
		case 2: [d exceptionEventOnStream: s]; break;
		default: abort ();
		}
	    }
    } while (!stop && [[TLDate now] compare: limit] < 0);

  changed = 1;
  stop = 0;

  GCUNPRO;

  return self;
} /* -runUntilDate: */

-setInactivityTimeout: (TLDate *) d
{
  ASGN_IVAR (inactive_timeout, d);
  return (self);
} /* -setInactivityTimeout: */

-setInactivityGCTimeLimit: (TLDate *) d
{
  ASGN_IVAR (inactive_gc_time_limit, d);
  return (self);
} /* -setInactivityGCTimeLimit: */

-setInactivityPartialThreshold: (id <TLNumber>) n
{
  inactive_gc_partial_threshold = [n unsignedLongValue];
  return (self);
} /* -setInactivityPartialThreshold: */

-setInactivityTotalThreshold: (id <TLNumber>) n
{
  inactive_gc_total_threshold = [n unsignedLongValue];
  return (self);
} /* -setInactivityTotalThreshold: */

-stop
{
  stop = 1;
  return (self);
} /* -stop */

/******************** garbage collection ********************/

-(void) gcReference
{
  int i, j;

  MARK (inactive_timeout);
  MARK (inactive_gc_time_limit);

  for (i = 0; i < 3; i++)
    for (j = 0; j < fds[i].num; j++)
      if (FD_ISSET (j, &fds[i].set))
	{
	  MARK (fds[i].streams[j]);
	  MARK (fds[i].delegates[j]);
	}
} /* -gcReference */

@end
