/* Implementation of TLTelnetStream class.
   This file is part of TL, Tiggr's Library.
   Written by Tiggr <tiggr@es.ele.tue.nl>
   Copyright (C) 1995, 1996 Pieter J. Schoenmakers
   TL is distributed WITHOUT ANY WARRANTY.
   See the file LICENSE in the TL distribution for details.

   $Id: TLTelnetStream.m,v 1.4 1996/01/23 12:18:42 tiggr Exp $  */

#define TLTELNETSTREAM_DECLARE_PRIVATE_METHODS
#import "tl/TLTelnetStream.h"
#import "tl/TLPatchedRoots.h"
#import "tl/TLString.h"
#import <string.h>

#define WILL 0
#define WONT 1
#define DO   2
#define DONT 3

@implementation TLTelnetStream

/******************** TLStream ********************/

-close
{
  return self;
} /* -close */

/******************** TLInputStream ********************/

-(int) readByte
{
  unsigned char c;
  int i;

  i = [self readBytes: 1 intoBuffer: &c];
  return (i == 1 ? c : TL_EOF);
} /* -readByte */

-(int) readBytes: (int) n intoBuffer: (char *) b
{
  int i, return_read, read = [stream readBytes: n intoBuffer: b];
  unsigned char *from, *to, *run, *end;

  /* Negotiation output buffer.  */
  unsigned char neg_buf[10];
  int neg_num;

  if (read <= 0)
    return (read);

  for (return_read = read, from = to = (void *) b, end = from + read;
       from < end; from++)
    switch (state)
      {
      case TELCS_DATA:
	switch (*from)
	  {
	  case TELCS_CR:
	    state = TELCS_CR;
	    return_read--;
	    break;

	  case TELCS_IAC:
	    state = TELCS_IAC;
	    return_read--;
	    break;

	  default:
	    if (subnegotiation)
	      {
		/* We don't do SB/SE yet, so we can't have told the other
                   side we WILL whatever it is trying to tell us now.  */
		abort ();
	      }
	    else
	      {
		/* Run up to the next CR or IAC.  */
		for (run = from + 1;
		     run < end && (*run != TELCS_CR && *run != TELCS_IAC);
		     run++);
		if (from != to)
		  memcpy (to, from, run - from);
		to += run - from;
		from = run - 1;
	      }
	    break;
	  }
	break;

      case TELCS_CR:
	switch (*from)
	  {
	  case TELCS_NUL:
	    *to++ = TELCS_CR;
	    break;

	  case TELCS_NL:
	    *to++ = TELCS_NL;
	    break;

	  default:
#if 0
	    /* Not while we're talking to *^&@%#(^&@$% VMS telnetd.  */
	    [self warning: "received `%c' (%d) after CR", *from, *from];
#endif
	    if (to != from)
	      {
		*to++ = TELCS_CR;
		return_read++;
	      }
	    else
	      [self warning: "UNHANDLED bad CR char at start: `%c' (%d)",
	       *from, *from];
	    *to++ = *from;
	    break;
	  }
	state = TELCS_DATA;
	break;

      case TELCS_IAC:
	switch (*from)
	  {
	  case TELCS_SE:
	    if (subnegotiation)
	      subnegotiation = 0;
	    else
	      [self warning: "spurious IAC SE (ignored)"];
	    state = TELCS_DATA;
	    break;

	  case TELCS_SB:
	    subnegotiation = 1;
	  case TELCS_WILL:
	  case TELCS_WONT:
	  case TELCS_DO:
	  case TELCS_DONT:
	    state = *from;
	    break;

	  case TELCS_IAC:
	    *to++ = *from;
	    state = TELCS_DATA;
	    return_read++;
	    break;
	  }
	return_read--;
	break;

      case TELCS_WILL:
      case TELCS_WONT:
      case TELCS_DO:
      case TELCS_DONT:
	{
	  static const unsigned char cs[2][4] =
	    {
	      {TELCS_WILL, TELCS_WONT, TELCS_DO, TELCS_DONT},
	      {TELCS_DO, TELCS_DONT, TELCS_WILL, TELCS_WONT},
	    };
	  int party = (state == TELCS_DO || state == TELCS_DONT);
	  enum tl_telnet_options option = *from;
	  int neg_reply = 0;

	  if (state == TELCS_WILL || state == TELCS_DO)
	    {
	      if (option >= TLTELNET_NUM_OPTIONS)
		{
		  /* An option we do not understand.  */
		  neg_buf[1] = cs[party][DONT];
		  neg_reply = 1;
		}
	      else switch (osm[party][option])
		{
		case TELOS_NO:
		  if (delegate
		      && (party ? [delegate telnetStream: self optionDo: option]
			  :  [delegate telnetStream: self optionWill: option]))
		    {
		      osm[party][option] = TELOS_YES;
		      neg_buf[1] = cs[party][DO];
		    }
		  else
		    {
		      /* XXX We could decide on some of the options
			 ourselves...  */
		      neg_buf[1] = cs[party][DONT];
		    }
		  neg_reply = 1;
		  break;

		case TELOS_WANTNO_EMPTY:
		case TELOS_WANTNO_OPPOSITE:
		  osm[party][option] = (osm[party][option] == TELOS_WANTNO_EMPTY
					? TELOS_NO : TELOS_YES);
		  if (party)
		    [self warning: "option %d: DO while WANTNO", option];
		  else
		    [self warning: "option %d: WILL while WANTNO", option];
		  break;

		case TELOS_WANTYES_EMPTY:
		  osm[party][option] = TELOS_YES;
		  break;

		case TELOS_WANTYES_OPPOSITE:
		  osm[party][option] = TELOS_WANTNO_EMPTY;
		  neg_buf[1] = cs[party][DONT];
		  neg_reply = 1;
		  break;

		case TELOS_YES:
		  break;

		default:
		  abort ();
		}
	    }
	  else /* state == TELCS_WONT || state == TELCS_DONT */
	    {
	      if (option >= TLTELNET_NUM_OPTIONS)
		{
		  /* An option we do not understand.  */
		}
	      else switch (osm[party][option])
		{
		case TELOS_NO:
		  break;

		case TELOS_WANTNO_EMPTY:
		  osm[party][option] = TELOS_NO;
		  break;

		case TELOS_WANTNO_OPPOSITE:
		  osm[party][option] = TELOS_WANTYES_EMPTY;
		  neg_buf[1] = cs[party][DO];
		  neg_reply = 1;
		  break;

		case TELOS_WANTYES_EMPTY:
		  osm[party][option] = TELOS_NO;
		  break;

		case TELOS_WANTYES_OPPOSITE:
		  osm[party][option] = TELOS_NO;
		  break;

		case TELOS_YES:
		  if (delegate)
		    if (party)
		      [delegate telnetStream: self optionDont: option];
		    else
		      [delegate telnetStream: self optionWont: option];
		  osm[party][option] = TELOS_NO;
		  neg_buf[1] = cs[party][DONT];
		  neg_reply = 1;
		  break;

		default:
		  abort ();
		}
	    }

	  if (neg_reply)
	    {
	      neg_buf[0] = TELCS_IAC;
	      neg_buf[2] = option;
	      neg_num = 3;

	      i = [stream writeBytes: neg_num fromBuffer: neg_buf];
	      if (i < 0)
		{
		  [self warning: "write negotiation %#: %s",
		   [TLString stringWithCString: neg_buf length: neg_num],
		   ERRMSG];
		  return (i);
		}
	      if (i != neg_num)
		abort ();
	    }

	  state = TELCS_DATA;
	  return_read--;
	}
	break;

      default:
	abort ();
	break;
      }
  return (return_read);
} /* -readBytes:intoBuffer: */

/******************** TLOutputStream ********************/

-(int) writeByte: (char) c
{
  int i = [self writeBytes: 1 fromBuffer: &c];
  return (i == 1 ? c : TL_EOF);
} /* -writeByte: */

-(int) writeBytes: (int) n fromBuffer: (const char *) b
{
  int i, j, m, written = 0;
  const unsigned char *ub = (void *) b;

  for (i = 0; i < n;)
    switch ((unsigned) ub[i])
      {
      case TELCS_CR:
	if ([stream writeBytes: 2 fromBuffer: "\r\0"] != 2)
	  /* XXX Or should there be a writing state for this case?  */
	  abort ();
	i++, written++;
	break;

      case TELCS_IAC:
	if ([stream writeBytes: 2 fromBuffer: "\377\377"] != 2)
	  /* XXX Or should there be a writing state for this case?  */
	  abort ();
	i++, written++;
	break;

      default:
	for (j = i + 1; j < n && ub[j] != TELCS_CR && ub[j] != TELCS_IAC; j++);
	m = [stream writeBytes: j - i fromBuffer: ub + i];
	if (m != j - i)
	  return (m < 0 ? m : written + m);
	written += m;
	i += m;
	break;
    }
  return (written);
} /* -writeBytes:fromBuffer: */

@end
