/* Interface to TLRunLoop class
   This file is part of TL, Tiggr's Library.
   Written by Tiggr <tiggr@es.ele.tue.nl>
   Copyright (C) 1995, 1996 Pieter J. Schoenmakers
   TL is distributed WITHOUT ANY WARRANTY.
   See the file LICENSE in the TL distribution for details.

   $Id: TLRunLoop.h,v 1.5 1996/01/23 12:19:22 tiggr Exp $  */

#import "TLObject.h"
#import "TLStream.h"
#import "TLDate.h"
#import "TLNumber.h"
#import <sys/types.h>
#import <time.h>

@protocol TLRunLoopReadDelegate
-readEventOnStream: (id <TLStream>) stream;
@end

@protocol TLRunLoopWriteDelegate
-writeEventOnStream: (id <TLStream>) stream;
@end

@protocol TLRunLoopExceptionDelegate
-exceptionEventOnStream: (id <TLStream>) stream;
@end

@protocol TLRunLoopDelegate <TLRunLoopReadDelegate, TLRunLoopWriteDelegate,
			     TLRunLoopExceptionDelegate>
@end

struct tlrunloop_fd_info
{
  /* The file descriptor bitset.  */
  fd_set set;

  /* The index of the highest bit set in SET, plus 1.  */
  int num;

  /* The information on the stream and the delegate for each filedescriptor.  */
  id <TLStream> *streams;
  id <TLRunLoopDelegate> *delegates;

  /* The capacity of the STREAMS and DELEGATES.  */
  int cap;
};

@interface TLRunLoop: TLObject
{
  /* Runloop information for each of the read, write and exception sets to
     select(2).  */
  struct tlrunloop_fd_info fds[3];

  /* The elapsed time without any event after which this runloop will invoke
     a GC run.  */
  TLDate *inactive_timeout;

  /* The time allowed during such a GC run.  If this is not set (which is
     the default), no GC will be done.  */
  TLDate *inactive_gc_time_limit;

  /* The minimum number of object allocations which must have been done
     since the last partial GC run before a GC run is initiated for this
     reason.  */
  unsigned long inactive_gc_partial_threshold;

  /* The minimum number of object allocations which must have been done
     since the last completed GC run before a GC run is initiated for this
     reason.  */
  unsigned long inactive_gc_total_threshold;

  /* !0 iff one of the sets has changed its NUM.  */
  unsigned int changed: 1;

  /* !0 iff we're to return from the most recent invocation of run.  */
  unsigned int stop: 1;
}

/******************** creation ********************/

/* Return a new, fresh, run loop.  */
+(TLRunLoop *) runLoop;

/* Return the shared runloop for this thread (XXX of which we only have
   one...).  */
+(TLRunLoop *) sharedRunLoop;

-init;

/******************** registration ********************/

/* See `-addStream:delegate:toSet:'.  */
-addStreamForRead: (id <TLStream>) s delegate: (id <TLRunLoopReadDelegate>) d;
-addStreamForWrite: (id <TLStream>) s delegate: (id <TLRunLoopWriteDelegate>) d;
-addStreamForException: (id <TLStream>) s
 delegate: (id <TLRunLoopExceptionDelegate>) d;

/* See `-removeStream:fromSet:'.  */
-removeStreamForRead: (id <TLStream>) s;
-removeStreamForWrite: (id <TLStream>) s;
-removeStreamForException: (id <TLStream>) s;

/******************** garbage collection control ********************/

-(TLDate *) inactivityTimeout;
-(TLDate *) inactivityGCTimeLimit;
-(unsigned long) inactivityPartialThreshold;
-(unsigned long) inactivityTotalThreshold;

-setInactivityTimeout: (TLDate *) d;
-setInactivityGCTimeLimit: (TLDate *) d;
-setInactivityPartialThreshold: (id <TLNumber>) n;
-setInactivityTotalThreshold: (id <TLNumber>) n;

/******************** running ********************/

/* Run until the end of the world.  */
-run;

/* Run until the date D, exclusive.  */
-runUntilDate: (TLDate *) d;

/* Limit the current running date to this very moment.  */
-stop;

#ifdef TLRUNLOOP_DECLARE_PRIVATE_METHODS

/* Add the STREAM and DELEGATE to the FDS at IDX.  Overrides existing
   information for the same file descriptor and index.  Returns the stream,
   or NIL if STREAM can not deliver a valid filedescriptor or IDX is out of
   range.  If DELEGATE is NIL, the STREAM is taken to be the delegate.  */
-addStream: (id <TLStream>) stream delegate: delegate toSet: (int) idx;

/* Remove a stream from the set at IDX.  */
-removeStream: (id <TLStream>) stream fromSet: (int) idx;

/******************** garbage collection ********************/

-(void) gcReference;

#endif

@end
